# PRG - Exam #

This is the assignment for the PRG exam. Follow the instructions precisely and please note:

- If the module executes any functions when imported (e.g. print), points will be subtracted.
- If the module fails to be imported (due to syntactic errors), the solution is invalid.

The modules (packages) you implement should be packed in a zip file (files should be in the root of
the zip file, zipping directly the files, not the containing folder) and uploaded to BRUTE. For
every module, class and function you implement, do not forget to add a docstring. Code quality
(mainly following naming conventions and writing docstrings) is part of the grading of your
solution.


## Task 1 [3p] ##

Create a module named "traceback_code" which will produce the following traceback when executed.

```
Traceback (most recent call last):
  File "traceback_code.py", line 6, in <module>
    a + b
TypeError: unsupported operand type(s) for +: 'int' and 'str'
```


## Task 2 [5p] ##

Create a module named "task2_functions" and implement three functions according to the unit tests
below. You can copy the unit tests to the same module.

```
import unittest

class TestFunctions(unittest.TestCase):

    def test_length_without_spaces(self):
        # [1p]
        self.assertEqual(8, length_without_spaces("I am robot"))
        self.assertEqual(0, length_without_spaces("     "))
        self.assertEqual(9, length_without_spaces("Excellent"))

    def test_deduplicate(self):
        # [1.5p]
        self.assertEqual([10, 2, 1], deduplicate([10, 2, 1, 2]))
        self.assertEqual([3, 1, 2], deduplicate([3, 1, 2]))
        self.assertEqual([3, 10, 2], deduplicate([3, 10, 10, 3, 2, 3, 2]))

    def test_letter_frequency(self):
        # [2.5p]
        self.assertEqual({"p": 1, "r": 2, "o": 1, "g": 2, "a": 1, "m": 2, "i": 1, "n": 1},
                         letter_frequency("programming"))
```


## Task 3 [1p] ##

Create a package with a single module containing the class definition below.

1. [1p] Choose the name of the package and module appropriately. Be sure to follow best practices.

```
class Payout:
    """A single payment on a payroll with a due date (string) and an amount in CZK (int)."""

    currency = "CZK"

    def __init__(self, due_date, amount):
        self.due_date = due_date
        self.amount = amount

    def __add__(self, other):
        if self.due_date != other.due_date:
            raise ValueError("Cannot combine payouts with different due dates")
        return Payout(self.due_date, self.amount + other.amount)

    def __repr__(self):
        return f"{self.due_date} {self.amount}"
```


## Task 3.1 [4p] ##

In the module created in Task 3, implement the following tasks.

1. [2p] Add support for comparing instances using the equality operator "==".
2. [2p] Write a unit test for the equality operator and put it in a test case named "TestPayout".


## Task 3.2 [5p] ##

In the module created in Task 3, implement the following tasks.

1. [2p] Write function "save_payout" which takes two arguments - a Payout instance and a file path,
    and will store the data from the instance under the path. The format of the file is up to you.
2. [1p] In function save_payout, if the file path already exists, raise an exception.
3. [2p] Write function "load_payout" which takes a single argument - a file path, and returns
    a Payout instance containing the data previously saved by save_payout.


## Task 3.3 [7p] ##

In the module created in Task 3, implement the following tasks.

1. [1p] Implement class Payroll that stores a payout sequence (sequence of Payout instances).
2. [2p] Add support for the plus operator, where the first argument is always a Payroll instance and
    the second argument can be either a Payout instance or also a Payroll instance. The result of
    the addition is a new Payroll instance.
3. [1p] Implement the equality operator - two Payroll instances are equal if and only if their
    payout sequences equal.
4. [3p] Implement class PartTimePayroll which inherits from Payroll and ensures that each
    payout in the sequence is less than 10000. Ensure this by raising an appropriate exception
    if not. Try to reuse as much code from Payroll as possible.

Payroll example:

```
>>> Payroll([Payout("2021-01-31", 35000)]) == Payroll([Payout("2021-01-31", 35000)])
True
>>> Payroll([Payout("2021-01-31", 35000)]) == Payroll([]) + Payout("2021-01-31", 35000)
True
```

## Epilogue ##

At the end of the exam, the following files and folders should be present in your zip file. They
should be in the root of the zip file, not in a nested folder.

- traceback_code.py - task 1
- task2_functions.py - task 2
- [folder of chosen name] - task 3
